/**
  *  \file client/scr-result.cc
  */

#include "client/scr-result.h"

#include "client/vcrplay.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/event.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "util/nls.h"
#include "flak/flak-db.h"
#include "flak/vis-ui.h"

/******************************** Playback *******************************/

// FIXME! This is almost cut&paste from playvcr.cc. It should probably be moved into a library.

static const UIKeyTranslator::Init VCR_SEL_KEYS[] = {
    { SDLK_UP,      cm_VCR_Prev },
    { SDLK_DOWN,    cm_VCR_Next },
    { vk_WheelDown, cm_VCR_Next },
    { vk_WheelUp,   cm_VCR_Prev },
    { vk_Quit,      cm_Escape }
};

/** Play VCRs. */
static void
playVcrs(GClassicVcrDatabase& db)
{
    ObjectHolder holder;
    UIWindow w(_("Visual Combat Recorder"), UIWindow::BLUE_WINDOW, COLOR_GRAY);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(holder.add(new UIButton(cm_VCR_Prev, "\036", '-')));
    but.add(holder.add(new UIButton(cm_VCR_Next, "\037", '+')));
    but.add(holder.add(new UIButton(cm_Enter, _("ENTER - Play"), SDLK_RETURN)));
    but.add(holder.add(new UISpacer()));
    but.add(holder.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE, UIButton::aStop)));

    WVcrSelector sel(db);
    w.setLayout(UIVBoxLayout::instance5);
    w.add(sel);
    w.add(but);
    w.add(holder.add(new UIKeyTranslator(VCR_SEL_KEYS, countof(VCR_SEL_KEYS))));
    w.pack();
    ui_root->centerWidget(w);
    ui_root->add(w);
    ui_root->run();
    ui_root->close(w);
}

class SimVcrVisitor : public GVcrDatabaseVisitor {
 public:
    void visitClassicDatabase(GClassicVcrDatabase& db)
        {
            playVcrs(db);
        }
    void visitFlakDatabase(GFlakVcrDatabase& db)
        {
            if (db.getCount())
                playFlakBattle(db.getBattle(0));
        }
};

/************************** WSimClassResultList **************************/

WSimClassResultList::WSimClassResultList(GPlayerSet players, GSimResultSummary& result)
    : UIListbox(0, 15, 40 * font_ems[FONT_NORMAL], 0),
      result(result),
      players(players),
      desc_width(0),
      column_width(0)
{
    setHeaderHeight(getItemHeight());
}

WSimClassResultList::~WSimClassResultList()
{ }

void
WSimClassResultList::drawPart(GfxCanvas& can, int from, int to)
{
    UISkin& sk = getSkin();
    sk.startDraw();

    GfxParameters& parm = sk;
    for (int i = from; i < to; ++i) {
        const int me = getTop() + i;
        if (me >= result.getNumClassResults()) {
            drawEntry(can, UIColor::tc_Static, i, string_t());
        } else {
            const UIColorSet& scheme = (me == getCurrent() ? BLACK_COLOR_SET : GRAY_COLOR_SET);
            const int y = i * getItemHeight() + getHeaderHeight() + extent.y;
            const GSimResultSummary::ClassResult& class_result = result.getClassResult(me);

            parm.bgcolor = standard_colors[scheme[UIColor::tc_Background]];
            parm.color = standard_colors[scheme[UIColor::tc_Static]];
            parm.font = fonts[FONT_NORMAL];
            parm.just_x = 0;
            parm.just_y = 0;
            outTextF(can, parm, extent.x, y, desc_width,
                     format("%d (%.1f%%)",
                            class_result.cumulative_weight,
                            100.0*class_result.cumulative_weight / result.getCumulativeWeight()));

            int x = extent.x + desc_width + column_width;
            parm.just_x = 2;
            for (int i = 1; i <= NUM_OWNERS; ++i) {
                if (players.contains(i)) {
                    if (class_result.owned_units[i] == 0)
                        outTextF(can, parm, x, y, column_width, "-");
                    else
                        outTextF(can, parm, x, y, column_width, numToString(class_result.owned_units[i]));
                    x += column_width;
                }
            }
            int xr = extent.x + extent.w;
            int xl = x - column_width;
            if (xl < xr)
                drawSolidBar(can, GfxRect(xl, y, xr - xl, font_heights[FONT_NORMAL]), parm.bgcolor);
        }
    }
}

void
WSimClassResultList::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UISkin& sk = getSkin();
    sk.startDraw();
    sk.getParameters().just_x = 0;
    sk.getParameters().just_y = 0;
    sk.setColor(UIColor::tc_Static);

    int x = extent.x + desc_width;
    for (int i = 1; i <= NUM_OWNERS; ++i) {
        if (players.contains(i)) {
            sk.outTextF(can, FONT_NORMAL, x, extent.y, column_width, getAdjName(i));
            x += column_width;
        }
    }

    UIListbox::draw(can, ui);
}

void
WSimClassResultList::setExtent(GfxRect newpos, bool by_layout)
{
    UIListbox::setExtent(newpos, by_layout);

    // count number of players
    int num_players = 0;
    for (int i = 1; i <= NUM_OWNERS; ++i)
        if (players.contains(i))
            ++num_players;

    // compute sizes
    // FIXME: PCC 1.x does this differently
    if (num_players != 0) {
        int raw_desc_width = fonts[FONT_NORMAL]->getTextWidth("999x (100.0%)") + 10;
        if (raw_desc_width > extent.w)
            raw_desc_width = extent.w;

        int raw_col_width = (extent.w - raw_desc_width) / num_players;
        if (raw_col_width < 1)
            raw_col_width = 1;
        if (raw_col_width > font_ems[FONT_NORMAL] * 7)
            raw_col_width = font_ems[FONT_NORMAL] * 7;

        column_width = raw_col_width;
        desc_width   = raw_desc_width; // extent.w - raw_col_width * num_players;
    }
}

void
WSimClassResultList::updateFromResult()
{
    setSize(result.getNumClassResults());
    adjustPosition();
    drawWidget();
    sig_moved.raise(getId());
}

/**************************** WSimResultWindow ***************************/

static GPlayerSet
countPlayers(const GSimState& st)
{
    GPlayerSet result;
    if (st.hasPlanet())
        result += st.getPlanet().getOwner();
    for (GSimState::ship_index_t i = 0; i < st.getNumShips(); ++i)
        result += st.getShip(i).getOwner();
    return result;
}

WSimResultWindow::WSimResultWindow(const GSimState& sim_setup, const GSimOptions& sim_options)
    : UIWindow(_("Simulation Results"), BLUE_WINDOW, COLOR_GRAY),
      sim_setup(sim_setup),
      sim_options(sim_options),
      last_result(),
      result_summary(),
      class_result_list(countPlayers(sim_setup), result_summary)
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& group_content = holder.add(new UIGroup(UIHBoxLayout::instance0));
    UIGroup& group_buttons = holder.add(new UIGroup(UIHBoxLayout::instance5));

    group_buttons.add(holder.add(new UIButton(_("ENTER - Watch sample"), SDLK_RETURN)));
    group_buttons.add(holder.add(new UIButton(cm_Escape, _("ESC"), SDLK_ESCAPE, UIButton::aStop)));
    group_buttons.add(holder.add(new UISpacer()));

    group_content.add(class_result_list);
    group_content.add(holder.add(new UIScrollbar(class_result_list)));

    add(group_content);
    add(group_buttons);

    pack();
}

WSimResultWindow::~WSimResultWindow()
{ }

/** Run simulation.
    \param series true if we're simulating a series; stop after one run
    \param max    maximum number of battles to generate, 0 for infinite

    \todo On SMT/Multicore machines, we could fork off some worker
    threads to do the runSimulation(). */
void
WSimResultWindow::runSimulation(bool series, int max)
{
    ProgressMonitor monitor;

    while (1) {
        GSimState new_state(sim_setup);
        last_result.init(result_summary.getNumBattles());
        ::runSimulation(new_state, sim_options, last_result, monitor);
        result_summary.addResult(sim_setup, new_state, last_result);

        /* check exit conditions */
        if (max != 0) {
            if (--max == 0)
                break;
        }
        if (series) {
            if (last_result.series_length != 0)
                if ((result_summary.getNumBattles() % last_result.series_length) == 0)
                    break;
        }

        /* poll events */
        UIEvent event;
        do
            getEvent(event, true);
        while (event.type != ev_Key && event.type != ev_Idle);
        if (event.type == ev_Key)
            break;
    }
}

/** Run first simulation.
    \return true on success, false if setup doesn't give any fights */
bool
WSimResultWindow::runFirstSimulation()
{
    ProgressMonitor monitor;

    GSimState new_state(sim_setup);
    last_result.init(result_summary.getNumBattles());
    ::runSimulation(new_state, sim_options, last_result, monitor);
    if (last_result.battles != 0 && last_result.battles->getCount() != 0) {
        result_summary.addResult(sim_setup, new_state, last_result);
        return true;
    } else {
        return false;
    }
}

bool
WSimResultWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_RETURN:
            SimVcrVisitor().visit(*result_summary.getClassResult(class_result_list.getCurrent()).sample_battle);
            return true;
         case ' ':
            runSimulation(false, 1);
            class_result_list.updateFromResult();
            return true;
         case 's':
            runSimulation(true, 0);
            class_result_list.updateFromResult();
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}
