/**
  *  \file game/planet.cc
  *  \brief Planet / Starbase Management
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/planet.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/player.h"
#include "game/ship.h"
#include "game/global.h"
#include "cpluslib/format.h"
#include "game/config.h"

static const TPlanet blank_planet = {
    UNKNOWNP,                               // owner
    UNKNOWNP,                               // planet_id
    { UNKNOWNC, UNKNOWNC, UNKNOWNC },       // fc
    UNKNOWNP, UNKNOWNP, UNKNOWNP,           // mi, fa, de
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ore mined
    UNKNOWNP,                               // colonists
    UNKNOWNP, UNKNOWNP,                     // sup, mc
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ground ore
    UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP, // ore density
    UNKNOWNP, UNKNOWNP,                     // ctax, ntax
    UNKNOWNS, UNKNOWNS,                     // chap, nhap
    UNKNOWNP, UNKNOWNP, UNKNOWNP,           // ngov, natives, nrace
    UNKNOWNP,                               // temp
    UNKNOWNP                                // base
};

GPlanet::FullInfo::FullInfo(int id, bool known)
{
    using std::fill_n;

    if (!known)
        data = blank_planet;
    data.planet_id = id;
    known_to_have_natives = false;
    mineral_time = colonist_time = native_time = cash_time = 0;
}

/******************************** GPlanet ********************************/

/** Constructor. Create blank planet of given Id.
    \param id Planet Id */
GPlanet::GPlanet(int id)
    : id(id), full_info(0), base_dat(0), x(0), y(0),
      known_to_not_exist(false),
      name("?"), flags(0), psrc_flags(), bsrc_flags()
{ }

GPlanet::~GPlanet()
{
    delete full_info;
    delete base_dat;
}

/** Create info record.
    \param known=true if the planet is known, and the caller is going to
                      fill in createInfo().data
    \param known=false to initialize the planet to all-unknown */
GPlanet::FullInfo&
GPlanet::createInfo(bool known)
{
    if (!full_info)
        full_info = new FullInfo(id, known);
    return *full_info;
}

/** Get data to be used for display.
    \returns planet data record */
const TPlanet&
GPlanet::getDisplayData() const
{
    return full_info ? full_info->data : blank_planet;
}

/** \name Loading */
//@{

/** Add .dat file entry. This assumes that when we see a planet
    through several .dat files, we get the same one each time.
    \param dat   parsed .dat file entry
    \param srcf  source flag value to use for this entry */
void
GPlanet::addPlanetData(const TPlanet& dat, GPlayerSet srcf)
{
    // FIXME: older PHost versions clear the following fields of a
    // planet when sending a pdata planet target for exploration of an
    // unowned planet (owned planets never generate a target):
    // colonists, supplies, credits, mines, factories, defense,
    // coltax, nattax. We may want to merge that information somehow,
    // or detect and/or merge it.
    createInfo(true).data = dat;
    psrc_flags |= srcf;
}

/** Add starbase .dat file entry. See addDatEntry() for more
    information. */
void
GPlanet::addBaseData(const TStarbase& bdat, GPlayerSet srcf)
{
    if (!base_dat)
        base_dat = new TStarbase(bdat);
    bsrc_flags |= srcf;
}

/** Set planet position. */
void
GPlanet::setXY(int ax, int ay)
{
    x = ax;
    y = ay;
}

/** Set planet name. */
void
GPlanet::setName(const string_t& aname)
{
    name = aname;
}

/** Set whether non-existance of this planet is known. There is no way
    to explicitly specify that a planet does not exist. To build maps
    with fewer than 500 planets, people move planets to odd positions.
    Recent PHosts send a util.dat message whenever they consider a
    planet to be non-existant, to make sure that the clients' idea of
    which planets do exist agrees with PHost's. */
void
GPlanet::setKnownToNotExist(bool value)
{
    known_to_not_exist = value;
}

//@}

//! \name Checking
//@{

/** Do internal checks for this planet. Internal checks do not require
    a partner to interact with. This will fix the problems, and
    display appropriate messages. */
void
GPlanet::internalCheck()
{
    flags = 0;

    /* set existance flag according to x/y */
    if (isClippedCoord(GPoint(x, y)) || known_to_not_exist) {
        flags &= ~PLANET_EXISTS;
    } else {
        flags |= PLANET_EXISTS;
    }

    if (havePlanetData()) {
        /* we have a DAT entry. This planet MUST exist. */
        if (!(flags & PLANET_EXISTS)) {
            const char* fmt = (known_to_not_exist
                               ? _("Planet #%d has data although it is reported as non-existant -- host confused?")
                               : _("Planet #%d exists for Host, but is outside valid range"));
            console.write(LOG_WARN, format(fmt, id));
            flags |= PLANET_EXISTS;
        }
    }

    if (haveBaseData()) {
        /* we have a BDATA entry. There must also be a PDATA entry */
        if (!havePlanetData()) {
            console.write(LOG_WARN,
                          format(_("Starbase #%d does not have a planet -- deleting it"), id));
            delete base_dat;
            base_dat = 0;
            bsrc_flags = GPlayerSet();
        }
    }
}

/** Combined checks, phase 2. This will do all post-processing which
    needs a partner to interact with. */
void
GPlanet::combinedCheck2(GUniverse& univ)
{
    if (havePlanetData()) {
        full_info->mineral_time
            = full_info->colonist_time
            = full_info->native_time
            = full_info->cash_time
            = univ.getTurnNumber();
    }
}
//@}

//! @name Accessors
//@{

/** True iff this planet has a starbase. */
bool
GPlanet::hasBase() const
{
    return base_dat != 0;
}

/** Notify planet of ship in orbit. This will compute the planet
    markings. */
void
GPlanet::setShipInOrbit(const GShip& s)
{
    switch (getPlayerRelation(s.getOwner())) {
     case is_Me:
        flags |= PLANET_OWNORBIT;
        break;
     case is_Enemy:
        flags |= PLANET_ENEMYORBIT;
        break;
     case is_Ally:
        flags |= PLANET_ALLYORBIT;
        break;
    }
}

/** Check whether specified coordinates are clipped (i.e.\ nonexistant
    planet). There is no way for clients to tell that a planet is
    non-existant; the ad-hoc solution is moving planets far away.
    Consistent with PCC 1.x handling of this. */
bool
GPlanet::isClippedCoord(GPoint pt)
{
    if (pt.x <= 0 || pt.y <= 0 || pt.x > 9000 || pt.y > 9000 || !pt.isOnMap())
        return true;
    else
        return false;
}

string_t
GPlanet::getName() const
{
    return name;
}

string_t
GPlanet::getFullName() const
{
    return format(_("%s (Planet #%d)"), name, id);
}

int
GPlanet::getId() const
{
    return id;
}

int
GPlanet::getOwner() const
{
    if (full_info)
        return full_info->data.owner;
    else
        return 0;
}

bool
GPlanet::isExistant() const
{
    return flags & PLANET_EXISTS;
}

GPoint
GPlanet::getPos() const
{
    return GPoint(x, y);
}

string_t
GPlanet::getFCode() const
{
    if (full_info)
        return getBasicString(full_info->data.fcode);
    else
        return string_t();
}

void
GPlanet::setFCode(const string_t& str)
{
    storeBasicString(createInfo(false).data.fcode, str);
    setIsDirty(ps_DirtyData);
}

bool
GPlanet::isFCodeKnown() const
{
    return full_info
        && (full_info->data.fcode[0] != UNKNOWNC
            || full_info->data.fcode[1] != UNKNOWNC
            || full_info->data.fcode[2] != UNKNOWNC);
}

// Colonists

mn16_t
GPlanet::getColonistHappiness() const
{
    return mn16_t(getDisplayData().colonist_happy);
}

void
GPlanet::setColonistHappiness(mn16_t happiness)
{
    createInfo(false).data.colonist_happy = happiness.getRawValue();
    setIsDirty(ps_DirtyData);
}

mp16_t
GPlanet::getColonistTax() const
{
    return mp16_t(getDisplayData().colonist_tax);
}

void
GPlanet::setColonistTax(mp16_t tax)
{
    createInfo(false).data.colonist_tax = tax.getRawValue();
    setIsDirty(ps_DirtyData);
}

mn32_t
GPlanet::getColonists() const
{
    // if we do not know the amount, assume the reservation is empty
    return mn32_t(getDisplayData().colonists - cargo.getReserved(el_Colonists));
}

void
GPlanet::setColonists(mn32_t clans)
{
    if (clans != getColonists())
        return;
    createInfo(false).data.colonists = clans.getRawValue();
    setIsDirty(ps_DirtyData);
}

// Natives

bool
GPlanet::isKnownToHaveNatives() const
{
    /* We claim the planet has natives if
       ...we know that it has some from RGA        -or-
       ...we know the race (bioscan)               -or-
       ...we know the population (pillage) */
    return full_info
        && (full_info->known_to_have_natives
            || (mp16_t(full_info->data.native_race).isKnown()
                && full_info->data.native_race != nr_None)
            || (mp32_t(full_info->data.natives).isKnown()
                && full_info->data.natives != 0));
}

mp16_t
GPlanet::getNativeTax() const
{
    return mp16_t(getDisplayData().native_tax);
}

void
GPlanet::setNativeTax(mp16_t tax)
{
    createInfo(false).data.native_tax = tax.getRawValue();
    setIsDirty(ps_DirtyData);
}

mp16_t
GPlanet::getNativeGovernment() const
{
    return mp16_t(getDisplayData().native_gov);
}

void
GPlanet::setNativeGovernment(mp16_t gov)
{
    createInfo(false).data.native_gov = gov.getRawValue();
    setIsDirty(ps_DirtyData);
}

mn16_t
GPlanet::getNativeHappiness() const
{
    return mn16_t(getDisplayData().native_happy);
}

void
GPlanet::setNativeHappiness(mn16_t happiness)
{
    createInfo(false).data.native_happy = happiness.getRawValue();
    setIsDirty(ps_DirtyData);
}

mp16_t
GPlanet::getNativeRace() const
{
    return mp16_t(getDisplayData().native_race);
}

void
GPlanet::setNativeRace(mp16_t race)
{
    createInfo(false).data.native_race = race.getRawValue();
    setIsDirty(ps_DirtyData);
}

mp32_t
GPlanet::getNatives() const
{
    return mp32_t(getDisplayData().natives);
}

void
GPlanet::setNatives(mp32_t natives)
{
    createInfo(false).data.natives = natives.getRawValue();
    setIsDirty(ps_DirtyData);
}

void
GPlanet::setKnownToHaveNatives(bool known)
{
    createInfo(false).known_to_have_natives = known;
    setIsDirty(ps_DirtyData);
}

void
GPlanet::setBuildBaseFlag(bool b)
{
    ASSERT(havePlanetData());
    if (b != isBuildingBase()) {
        full_info->data.build_base = b;
        setIsDirty(ps_DirtyData);
    }
}

// Environment

mp16_t
GPlanet::getOreDensity(int index) const
{
    return mp16_t(getDisplayData().ore_density[index]);
}

void
GPlanet::setOreDensity(int index, mp16_t amount)
{
    createInfo(false).data.ore_density[index] = amount;
    setIsDirty(ps_DirtyData);
}

mp32_t
GPlanet::getOreGround(int index) const
{
    return mp32_t(getDisplayData().ore_ground[index]);
}

void
GPlanet::setOreGround(int index, mp32_t amount)
{
    createInfo(false).data.ore_ground[index] = amount.getRawValue();
    setIsDirty(ps_DirtyData);
}

mp16_t
GPlanet::getTemp() const
{
    mp16_t temp_code = mp16_t(getDisplayData().temp_code);
    if (temp_code.isKnown())
        return mp16_t(100 - temp_code);
    else
        return temp_code;
}

void
GPlanet::setTemp(mp16_t temp)
{
    if (temp.isKnown())
        createInfo(false).data.temp_code = 100 - temp;
    else if (full_info)
        full_info->data.temp_code = temp.getRawValue();
    setIsDirty(ps_DirtyData);
}

/** Get cargo of a particular type, as to be used by most code.
    This will remove reservations from the amount reported. */
int32
GPlanet::getCargo(GCargoType type) const
{
    return getCargoRaw(type) - cargo.getReserved(type);
}

/** Get cargo, as it is in underlying data structure. */
int32
GPlanet::getCargoRaw(GCargoType type) const
{
    const TPlanet& dpy = getDisplayData();
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return dpy.ore_mined[type];
     case el_Supplies:
        return dpy.supplies;
     case el_Money:
        return dpy.money;
     case el_Fighters:
        if (base_dat)
            return base_dat->ammo_store[10];
        return 0;
     case el_Colonists:
        return dpy.colonists;
     default:
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        if (base_dat)
            return base_dat->ammo_store[type - el_Torps];
        return 0;
    }
}

void
GPlanet::changeCargoRaw(GCargoType type, int32 amount)
{
    if (!amount)
        return;

    // FIXME
    ASSERT(havePlanetData());
    setIsDirty(ps_DirtyData);

    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        full_info->data.ore_mined[type] += amount;
        return;
     case el_Supplies:
        full_info->data.supplies += amount;
        return;
     case el_Money:
        full_info->data.money += amount;
        return;
     case el_Fighters:
        ASSERT(base_dat);
        base_dat->ammo_store[10] += amount;
        return;
     case el_Colonists:
        full_info->data.colonists += amount;
        return;
     default:
        ASSERT(base_dat);
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        base_dat->ammo_store[type - el_Torps] += amount;
        return;
    }
}

// FIXME: this function should go
bool
GPlanet::isCargoAmountKnown(GCargoType type) const
{
    if (!full_info)
        return false;
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return full_info->data.ore_mined[type] != UNKNOWNP;
     case el_Supplies:
        return full_info->data.supplies != UNKNOWNP;
     case el_Money:
        return full_info->data.money != UNKNOWNP;
     case el_Fighters:
        return base_dat != 0;
     case el_Colonists:
        return full_info->data.colonists != UNKNOWNP;
     default:
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        return base_dat != 0;
    }
}

mp16_t
GPlanet::getStructures(TPlanetaryStructure ps) const
{
    if (!full_info)
        return mp16_t();

    switch (ps) {
     case str_Mines:
        return mp16_t(full_info->data.mines);
     case str_Factories:
        return mp16_t(full_info->data.factories);
     case str_Defense:
        return mp16_t(full_info->data.defense);
     case str_BaseDefense:
        // FIXME: if we do not know whether there is a base, the right
        // answer would be mp16_t().
        return base_dat ? mp16_t(base_dat->defense) : mp16_t(0);
    }
    return mp16_t();
}

void
GPlanet::setStructures(TPlanetaryStructure ps, mp16_t count)
{
    // FIXME: does not handle that we know info about foreign planet
    if (count == getStructures(ps))
        return;
    ASSERT(havePlanetData());
    switch (ps) {
     case str_Mines:
        full_info->data.mines = count.getRawValue();
        break;
     case str_Factories:
        full_info->data.factories = count.getRawValue();
        break;
     case str_Defense:
        full_info->data.defense = count.getRawValue();
        break;
     case str_BaseDefense:
        ASSERT(base_dat);
        base_dat->defense = count.getRawValue();
        break;
    }
    setIsDirty(ps_DirtyData);
}

void
GPlanet::setBaseMission(int mission)
{
    ASSERT(base_dat);
    if (mission != getBaseMission()) {
        base_dat->mission = mission;
        setIsDirty(ps_DirtyData);
    }
}

void
GPlanet::setBaseShipyardOrder(int action, int id)
{
    ASSERT(base_dat);
    if (action != getBaseShipyardOrder() || id != getBaseShipyardId()) {
        base_dat->shipyard_action = action;
        base_dat->shipyard_id     = id;
        setIsDirty(ps_DirtyData);
    }
}

void
GPlanet::setBaseTech(TTechLevel level, int value)
{
    ASSERT(value > 0 && value <= 10);
    ASSERT(base_dat);
    if (value != base_dat->tech_levels[level]) {
        base_dat->tech_levels[level] = value;
        setIsDirty(ps_DirtyData);
    }
}

int
GPlanet::getBaseBuildHull() const
{
    ASSERT(base_dat);
    if (base_dat->build.hull)
        return getTruehull(base_dat->owner, base_dat->build.hull);
    else
        return 0;
}

TBuildOrder
GPlanet::getBaseBuildOrder() const
{
    TBuildOrder bo = base_dat->build;
    bo.hull = getBaseBuildHull();
    return bo;
}

void
GPlanet::setBaseBuildOrder(const TBuildOrder& order)
{
    if (!order.hull) {
        base_dat->build.hull = base_dat->build.engine =
            base_dat->build.beam_type = base_dat->build.beam_count =
            base_dat->build.torp_type = base_dat->build.torp_count =
            base_dat->build.unused = 0;
    } else {
        int slot = getTruehullSlot(base_dat->owner, order.hull);
        ASSERT(slot);
        base_dat->build = order;
        base_dat->build.hull = slot;
    }
    setIsDirty(ps_DirtyData);
}

/** Set number of engines in starbase storage. */
void
GPlanet::setBaseEngineStore(int slot, int amount)
{
    ASSERT(slot > 0 && slot <= NUM_ENGINES);
    base_dat->engine_store[slot-1] = amount;
    setIsDirty(ps_DirtyData);
}

/** Set number of beams in starbase storage. */
void
GPlanet::setBaseBeamStore(int slot, int amount)
{
    ASSERT(slot > 0 && slot <= NUM_BEAMS);
    base_dat->beam_store[slot-1] = amount;
    setIsDirty(ps_DirtyData);
}

/** Set number of hulls in starbase storage. Note that this takes a
    truehull slot number, not a real hull Id. */
void
GPlanet::setBaseHullStoreSlot(int slot, int amount)
{
    ASSERT(slot > 0 && slot <= NUM_HULLS_PER_PLAYER);
    base_dat->hull_store[slot-1] = amount;
    setIsDirty(ps_DirtyData);
}

/** Set number of torpedo launchers in starbase storage. */
void
GPlanet::setBaseLauncherStore(int slot, int amount)
{
    ASSERT(slot > 0 && slot <= NUM_TORPS);
    base_dat->launcher_store[slot-1] = amount;
    setIsDirty(ps_DirtyData);
}

/** Get ship which is trying to clone at this planet. This runs
    through the ship list, so use with care.
    \param univ universe which contains the ships */
int
GPlanet::getShipCloningHere(const GUniverse& univ) const
{
    // FIXME: check whether 'cln' works
    if (!config.AllowShipCloning)
        return 0;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& s = univ.getShip(i);
        if (s.isExistant() && s.isCloningAt(*this))
            return i;
    }
    return 0;
}

/****************************** GPlanetType ******************************/

/** Constructor.
    \param trn Turn object through which we iterate. */
GPlanetType::GPlanetType(GUniverse& univ)
    : GObjectType(NUM_PLANETS, univ)
{ }

bool
GPlanetType::isExistant(int index)
{
    if (index > 0 && index <= NUM_PLANETS) {
        const GPlanet& p = getUniverse().getPlanet(index);
        if (p.isExistant() && p.isPlayed() && p.havePlanetData())
            return true;
    }
    return false;
}

GPlanet&
GPlanetType::getObject(int index)
{
    return getUniverse().getPlanet(index);
}

string_t
GPlanetType::getClassName()
{
    return string_t(_("Planet"));
}

/******************************* GBaseType *******************************/

GBaseType::GBaseType(GUniverse& univ)
    : GObjectType(NUM_PLANETS, univ)
{ }

bool
GBaseType::isExistant(int index)
{
    if (index > 0 && index <= NUM_PLANETS) {
        const GPlanet& p = getUniverse().getPlanet(index);
        if (p.isExistant() && p.isPlayed() && p.hasBase())
            return true;
    }
    return false;
}

GPlanet&
GBaseType::getObject(int index)
{
    return getUniverse().getPlanet(index);
}

string_t
GBaseType::getClassName()
{
    return string_t(_("Starbase"));
}

/**************************** GPlanetTransfer ****************************/

GPlanetTransfer::GPlanetTransfer(GPlanet& planet)
    : planet(planet)
{
    planet.getCargoHandler().incUseCount();
}

GPlanetTransfer::~GPlanetTransfer()
{
    planet.getCargoHandler().decUseCount();
}

int32
GPlanetTransfer::getAmount(GCargoType type)
{
    return planet.getCargo(type);
}

void
GPlanetTransfer::changeAmount(GCargoType type, int32 amount)
{
    planet.changeCargoRaw(type, amount);
}

void
GPlanetTransfer::reserveAmount(GCargoType type, int32 amount)
{
    planet.getCargoHandler().changeReservation(type, amount);
}

bool
GPlanetTransfer::canHave(GCargoType cargo_id)
{
    if (cargo_id == el_Fighters
        || (cargo_id >= el_Torps && cargo_id < el_Torps + NUM_TORPS))
    {
        if (!planet.hasBase())
            return false;
    }
    return true;
}

int32
GPlanetTransfer::getMax(GCargoType cargo_id)
{
    // FIXME: Host .31 torp safety
    /* the torp/ore limits are arbitrary, but should match Tim's TRN check. */
    if (cargo_id == el_Fighters)
        return 60;
    else if (cargo_id >= el_Torps && cargo_id < el_Torps + NUM_TORPS)
        return 10000;
    else
        return 1000000000;
}

string_t
GPlanetTransfer::getName() const
{
    // FIXME: PCC 1.x says "$NAME starbase" or "planet $NAME" here.
    // should we use same wording?
    return planet.getName();
}

//@}
