/**
  *  \file game/transact.h
  *  \brief Transactions
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_TRANSACT_H
#define PCC_V2_GAME_TRANSACT_H

#include "cpluslib/sigslot.h"
#include "cpluslib/holder.h"
#include "game/error.h"

class GMultiTransaction;

//! Transaction.
class GTransaction : public Object {
    friend class GMultiTransaction;

    // Result tracking
    enum Result { Nothing, Committed, Cancelled };
    Result result      : 8;
    Result must_result : 8;
    void finish(Result);

    // Multi-Part Transactions
    GTransaction* next;
    GMultiTransaction* parent;
 protected:
    /** Constructor. */
    GTransaction()
        : result(Nothing), must_result(Nothing), next(0), parent(0) { }

    /** Commit the transaction. Derived classes must override this
        method to do the actual work. By convention, a transaction
        should not commit other transactions it did not create
        itself. Instead, the transaction user should use a
        GMultiTransaction to set them up as a group and commit them
        as a group. For consistency checking, you can call
        mustCommit() on the related transactions to make sure they
        are committed, not cancelled.
        \pre !getError() */
    virtual void commit() = 0;

    /** Cancel (rollback) the transaction. Derived classes must
        override this method to do the actual rollback. The convention
        described for commit() applies here, too. You can use
        mustCancel() to mark a transaction as needing to be
        cancelled. */
    virtual void cancel() = 0;
public:
    virtual ~GTransaction();

    void doCommit();
    void doCancel();

    /** Check whether this transaction was finished cleanly (commit or
        cancel). */
    bool isFinished() const
        { return result != Nothing; }
    /** Check whether this transaction was committed (i.e.\ it is
        neither unfinished nor cancelled). */
    bool isCommitted() const
        { return result == Committed; }
    /** Check whether this transaction was cancelled (i.e.\ it is
        neither unfinished nor committed). */
    bool isCancelled() const
        { return result == Cancelled; }

    void mustCommit();
    void mustCancel();

    virtual GError getError();

    /** Change signal. The transaction implementation should call this
        when something related to the transaction changes, and update
        of the display may be required. This is provided as a
        convenience; GTransaction does not use it. */
    Signal0 sig_changed;
};

//! Multi-Part Transaction.
class GMultiTransaction : public GTransaction {
    GTransaction* first;
    void doAddNew(GTransaction*);

 protected:
    void commit();
    void cancel();

 public:
    GMultiTransaction();
    ~GMultiTransaction();

    /** Add newly-allocated subtransaction. Use as
        <tt>MyTA& m = mpt.addNewSubtransaction(new MyTA())</tt>. */
    template<class T>
    T& addNewSubtransaction(T* t)
        { doAddNew(t); return *t; }

    void addSubtransaction(GTransaction& ta);
    void removeSubtransaction(GTransaction& ta);
    void removeAllSubtransactions();

    GError getError();

    ObjectHolder holder;
};

#endif
