/**
  *  \file game/univ.h
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_UNIV_H
#define PCC_V2_GAME_UNIV_H

#include <vector>
#include <map>
#include "cpluslib/assert.h"
#include "cpluslib/string.h"
#include "game/coord.h"
#include "game/struct.h"
#include "game/player.h"
#include "cpluslib/sigslot.h"

class GPlanet;
class GPlanetType;
class GBaseType;
class GShip;
class GShipType;
class GIonStorm;
class GIonStormType;
class GMinefield;

//! Flags for GUniverse::getLocationName().
enum {
    locs_Orbit   = 1,    ///< Show "Orbit of" for planet names.
    locs_Verbose = 2,    ///< Be more verbose.
    locs_Ships   = 4,    ///< Show a ship name if applicable.
    locs_WW      = 8     ///< Show planet name if point is in warp well.
};

class GUniverse {
 public:
    typedef std::map<int,GMinefield*> MinefieldMap;

    GUniverse();
    ~GUniverse();

    // Initialisation
    void init();
    void done();
    void postprocess(GPlayerSet playing_set);

    // Turn number
    void setTurnNumber(int nr);
    int  getTurnNumber() const;

    // Object accessors
    GPlanet&          getPlanet(int id);
    const GPlanet&    getPlanet(int id) const;
    GShip&            getShip(int id);
    const GShip&      getShip(int id) const;
    GIonStorm&        getIonStorm(int sid);
    const GIonStorm&  getIonStorm(int sid) const;
    GMinefield*       getMinefield(int id, int cause = -1);

    MinefieldMap::iterator getMinefieldBegin()
        { return minefields.begin(); }
    MinefieldMap::const_iterator getMinefieldBegin() const
        { return minefields.begin(); }
    MinefieldMap::iterator getMinefieldEnd()
        { return minefields.end(); }
    MinefieldMap::const_iterator getMinefieldEnd() const
        { return minefields.end(); }

    // Type accessors
    GShipType& getShipType()
        { return *ship_type; }
    GPlanetType& getPlanetType()
        { return *planet_type; }
    GBaseType& getBaseType()
        { return *base_type; }
    GIonStormType& getIonStormType()
        { return *storm_type; }

    // Location accessors
    int       getPlanetAt(GPoint pt) const;
    int       getPlanetAt(GPoint pt, bool gravity) const;
    int       getGravityPlanetAt(GPoint pt) const;
    int       getAnyShipAt(GPoint pt) const;
    int       getNumberOfPlayedShipsAt(GPoint pt) const;
    string_t  getLocationName(GPoint pt, int flags) const;

    // Updates
    void doUpdates();

    /// Signal: Planet was changed. Parameter is planet Id.
    Signal1<int> sig_planet_changed;
    /// Signal: Ship was changed. Parameter is ship Id.
    Signal1<int> sig_ship_changed;
 private:
    GPlanet*     planets[NUM_PLANETS];
    GShip*       ships[NUM_SHIPS];
    GIonStorm*   storms[NUM_STORMS];
    MinefieldMap minefields;

    GShipType*     ship_type;
    GPlanetType*   planet_type;
    GBaseType*     base_type;
    GIonStormType* storm_type;

    int turn_nr;

    void clearAllMines();
};

inline void
GUniverse::setTurnNumber(int nr)
{
    turn_nr = nr;
}

inline int
GUniverse::getTurnNumber() const
{
    return turn_nr;
}

#endif
